@@ -1,4 +1,4 @@
-import fakeredis, sys, math, json, redis, time, random, itertools, datetime
+import fakeredis, sys, math, json, redis, time, random, itertools, datetime, muid
 import numpy as np
 from collections import Counter
 from typing import List, Union, Any, Optional
@@ -17,8 +17,6 @@ PY_REDIS_ARGS = ('host','port','db','username','password','socket_timeout','sock
                  'ssl_check_hostname', 'max_connections', 'single_connection_client','health_check_interval', 'client_name')
 FAKE_REDIS_ARGS = ('decode_responses',)
 
-
-
 class Rediz(RedizConventions):
 
     # Initialization
@@ -69,6 +67,9 @@ class Rediz(RedizConventions):
     def get_samples(self, name, delay=None, delays=None):
         return self._get_samples_implementation(name=name, delay=delay, delays=delays)
 
+    def get_index(self):
+        return self._get_index_implementation()
+
     def get_predictions(self, name, delay=None, delays=None):
         return self._get_predictions_implementation(name=name, delay=delay, delays=delays)
 
@@ -92,14 +93,11 @@ class Rediz(RedizConventions):
     def get_lagged_times(self, name, start=0, end=None, count=None, to_float=True):
         return self._get_lagged_implementation(name, start=start, end=end, count=count, with_values=False, with_times=True, to_float=to_float)
 
-    def get_balance(self, write_key ):
-        return self._get_balance_implementation(write_key=write_key)
+    def get_performance(self, write_key, name=None, delay=None):
+        return self._get_performance_implementation(write_key=write_key, name=name, delay=delay)
 
-    def mget_balance(self, write_keys, aggregate=True):
-        return self._get_balance_implementation(write_keys=write_keys, aggregate=aggregate)
-
-    def get_performance(self, name, delay=None, count=50):
-        return self._get_performance_implementation(name=name, delay=delay, count=count )
+    def get_leaderboard(self, name, delay=None, count=50):
+        return self._get_leaderboard_implementation(name=name, delay=delay, count=count)
 
     def get_history(self, name, max='+', min='-', count=None, populate=True, drop_expired=True ):
         return self._get_history_implementation( name=name, max=max, min=min, count=count, populate=populate, drop_expired=drop_expired )
@@ -113,6 +111,15 @@ class Rediz(RedizConventions):
     def get_errors(self, write_key, start=0, end=-1):
         return self.client.lrange(name=self.errors_name(write_key=write_key), start=start, end=end)
 
+    def get_confirms(self, write_key, start=0, end=-1):
+        return self.client.lrange(name=self.confirms_name(write_key=write_key), start=start, end=end)
+
+    def get_balance(self, write_key):
+        return self.client.hget(name=self._BALANCES, key=write_key)
+
+    def get_performance(self, write_key):
+        return self.client.hgetall(name=self.performance_name(write_key=write_key))
+
     def get_links(self, name, delay=None, delays=None ):
         assert not self.SEP in name, "Intent is to provide delay variable"
         return self._get_links_implementation(name=name, delay=delay, delays=delays )
@@ -120,8 +127,8 @@ class Rediz(RedizConventions):
     def get_backlinks(self, name ):
         return self._get_backlinks_implementation(name=name )
 
-    def get_transactions(self, max='+', min='-', count=None, write_key=None, name=None):
-        return self._get_transactions_implementation(max=max,min=min,count=count, write_key=write_key, name=name)
+    def get_transactions(self, max='+', min='-', count=None, write_key=None, name=None, delay=None):
+        return self._get_transactions_implementation(max=max,min=min,count=count, write_key=write_key, name=name, delay=delay)
 
     def get_summary(self,name):
         assert self._root_name(name)==name
@@ -149,6 +156,7 @@ class Rediz(RedizConventions):
     def set( self, name, value, write_key, budget=10 ):
         """ Set name=value and initiate clearing, derived zscore market etc """
         assert RedizConventions.is_plain_name(name),"Expecting plain name"
+        assert RedizConventions.is_valid_key(write_key),"Invalid write_key"
         return self._set_implementation(name=name, value=value, write_key=write_key, return_args=None, budget=budget, with_percentiles=True )
 
     def mset(self,names:NameList, values:ValueList, write_keys:KeyList, budgets:List[int] ):
@@ -171,9 +179,9 @@ class Rediz(RedizConventions):
 
     def set_scenarios(self, name, values, delay, write_key):
         """ Supply scenarios for scalar value taken by name
-               values :   [ float ]  len  self.NUM_PREDICTIONS
+               values :   [ float ]  len  self.num_predictions
         """
-        assert len(values)==self.NUM_PREDICTIONS
+        assert len(values)==self.num_predictions
         assert delay in self.DELAYS
         assert self.is_valid_key(write_key)
         fvalues = list(map(float,values))
@@ -216,7 +224,6 @@ class Rediz(RedizConventions):
         """ Permissioned removal of link (either party can do this) """
         return self._unlink_implementation(name=name, delay=delay, write_key=write_key, target=target )
 
-
     # --------------------------------------------------------------------------
     #            Implementation  (client init)
     # --------------------------------------------------------------------------
@@ -260,7 +267,6 @@ class Rediz(RedizConventions):
         names = list_or_args(names,args)
         return self.client.hmget(self._ownership_name(),*names)
 
-
     # --------------------------------------------------------------------------
     #            Implementation  (set)
     # --------------------------------------------------------------------------
@@ -312,6 +318,15 @@ class Rediz(RedizConventions):
             for title in titles:
                 if title["name"] in prctls:
                     title.update( {"percentiles":prctls[title["name"]]} )
+
+        # Write to confirmation log
+        cnfrms = self.confirms_name(write_key=write_keys[0])
+        confirm_pipe = self.client.pipeline(transaction=False)
+        confirm_pipe.lpush(cnfrms, json.dumps(titles[:10]))
+        confirm_pipe.expire(cnfrms, self.CONFIRMS_TTL)
+        confirm_pipe.ltrim(cnfrms, start=0, end=self.CONFIRMS_MAX)
+        confirm_pipe.execute(raise_on_error=True)
+
         return titles[0] if singular else titles
 
 
@@ -350,8 +365,6 @@ class Rediz(RedizConventions):
                     rejected.append({"ndx":ndx, "name":name,"write_key":None,"value":value,"error":"invalid value of type "+str(type(value))+" was supplied"})
                 else:
                     if name is None:
-                        if write_key is None:
-                            write_key = self.random_key()
                         if not(self.is_valid_key(write_key)):
                             rejected.append({"ndx":ndx,"name":name,"write_key":None,"errror":"invalid write_key"})
                         else:
@@ -392,8 +405,6 @@ class Rediz(RedizConventions):
             new_pipe     = self.client.pipeline(transaction=False)
             for exist, ndx, name, value, write_key, budget in zip( exists, ndxs, names, values, write_keys, budgets):
                 if not(exist):
-                    if write_key is None:
-                        write_key = self.random_key()
                     if not(self.is_valid_key(write_key)):
                         rejected.append({"ndx":ndx,"name":name,"write_key":None,"errror":"invalid write_key"})
                     else:
@@ -496,7 +507,7 @@ class Rediz(RedizConventions):
         # (1) Set the actual value ... which will be overwritten by the next set() ... and a randomly named copy that survives longer
         ttl = self._cost_based_ttl(value=value,budget=budget)
         pipe.set(name=name,value=value,ex=ttl)
-        name_of_copy = self._promised_name(name)
+        name_of_copy = self._random_promised_name(name)
         promise_ttl = self._promise_ttl()
         pipe.set(name=name_of_copy, value=value, ex=promise_ttl)
 
@@ -855,7 +866,7 @@ class Rediz(RedizConventions):
          for value, destination, method in zip(source_values, destinations, methods):
              if method == 'copy':
                  if value is None:
-                     report['warnings'].append('None value found ')
+                     report['warnings'] = report['warnings'] + ' None value found '
                  else:
                      delay_ttl = int(max(self.DELAYS)+self._DELAY_GRACE+5*60)
                      move_pipe.set(name=destination,value=value,ex=delay_ttl)
@@ -889,7 +900,7 @@ class Rediz(RedizConventions):
 
     def _baseline_prediction(self, name, value, write_key, delay ):
         # As a finer point, we should really be using the delay times here and sampling by time not lag ... but it is just a lazy benchmark anyway
-        lagged_values = self._get_lagged_implementation(name, with_times=False, with_values=True, to_float=True, start=0, end=None, count=self.NUM_PREDICTIONS)
+        lagged_values = self._get_lagged_implementation(name, with_times=False, with_values=True, to_float=True, start=0, end=None, count=self.num_predictions)
         predictions = self.empirical_predictions(lagged_values=lagged_values)
         return self._set_scenarios_implementation(name=name, values=predictions, write_key=write_key, delay=delay)
 
@@ -903,7 +914,7 @@ class Rediz(RedizConventions):
             delete_pipe = self.client.pipeline(transaction=True)  # <-- Important that transaction=True
             for delay in delays:
                 collective_predictions_name = self._predictions_name(name, delay)
-                keys = [ self._format_scenario(self, write_key, k) for k in range(self.NUM_PREDICTIONS) ]
+                keys = [ self._format_scenario(self, write_key, k) for k in range(self.num_predictions) ]
                 delete_pipe.zrem( collective_predictions_name, *keys)
                 samples_name = self._samples_name(name=name, delay=delay)
                 delete_pipe.zrem(samples_name, *keys)
@@ -916,7 +927,7 @@ class Rediz(RedizConventions):
         """ Charge for this! Not encouraged as it should not be necessary, and it is inefficient to get scenarios back from the collective zset """
         assert name == self._root_name(name)
         if self.is_valid_key(write_key) and delay in self.DELAYS:
-            cursor, items = self.client.zscan(name=self._predictions_name(name=name, delay=delay),cursor=cursor,match='*'+write_key+'*',count=self.NUM_PREDICTIONS)
+            cursor, items = self.client.zscan(name=self._predictions_name(name=name, delay=delay),cursor=cursor,match='*'+write_key+'*',count=self.num_predictions)
             return {"cursor":cursor, "scenarios":dict(items)}
 
     def _get_invcdf_implementation(self, name, delay, percentiles):
@@ -929,14 +940,17 @@ class Rediz(RedizConventions):
         assert delay in self.DELAYS
         score_pipe = self.client.pipeline()
         num = self.client.zcard(name=self._predictions_name(name=name, delay=delay))
-        h = max( 100.0/num, 0.00001)*max([ abs(v) for v in values ]+[1.0])
-        for value in values:
-            score_pipe.zrangebyscore(name=self._predictions_name(name=name,delay=delay),min=value, max=value+h, start=0, num=10, withscores=False)
-        exec = score_pipe.execute()
-        #
-        prtcls = [ self._zmean_scenarios_percentile(percentile_scenarios=ex) if ex else np.NaN for ex in exec]
-        valid  = [ (v,p) for v,p in zip(values,prtcls) if not np.isnan(p) ]
-        return {"x":[v for v,p in valid], "y":[p for v,p in valid]}
+        if num:
+            h = max( 100.0/num, 0.00001)*max([ abs(v) for v in values ]+[1.0])
+            for value in values:
+                score_pipe.zrangebyscore(name=self._predictions_name(name=name,delay=delay),min=value, max=value+h, start=0, num=10, withscores=False)
+            exec = score_pipe.execute()
+            #
+            prtcls = [ self._zmean_scenarios_percentile(percentile_scenarios=ex) if ex else np.NaN for ex in exec]
+            valid  = [ (v,p) for v,p in zip(values,prtcls) if not np.isnan(p) ]
+            return {"x":[v for v,p in valid], "y":[p for v,p in valid]}
+        else:
+            return {"message":"No predictions."}
 
 
     def _set_scenarios_implementation(self, name, values, write_key, delay=None, delays=None):
@@ -946,12 +960,13 @@ class Rediz(RedizConventions):
         elif delays is None:
             delays = [ delay ]
         assert name==self._root_name(name)
-        if len(values)==self.NUM_PREDICTIONS and self.is_valid_key(write_key
+        if len(values)==self.num_predictions and self.is_valid_key(write_key
                 ) and all( [ isinstance(v,(int,float) ) for v in values] ) and all (delay in self.DELAYS for delay in delays):
             # Jigger sorted predictions
-            noise =  np.random.randn(self.NUM_PREDICTIONS).tolist()
+            noise =  np.random.randn(self.num_predictions).tolist()
             jiggered_values = [v + n*self.NOISE for v, n in zip(values, noise)]
             jiggered_values.sort()
+            assert len(set(jiggered_values))==len(jiggered_values),"coincidence??"
             predictions = dict([(self._format_scenario(write_key=write_key, k=k), v) for k, v in enumerate(jiggered_values)])
 
             # Open pipeline
@@ -960,11 +975,12 @@ class Rediz(RedizConventions):
             # Add to collective contemporaneous forward predictions
             for delay in delays:
                 collective_predictions_name = self._predictions_name(name, delay)
-                set_and_expire_pipe.zadd(   name=collective_predictions_name, mapping=predictions, ch=True,nx=False)  # (0)
+                set_and_expire_pipe.zadd(  name=collective_predictions_name, mapping=predictions, ch=True,nx=False)  # (0)
 
             # Create obscure predictions and promise to insert them later, at different times, into different samples
             utc_epoch_now = int(time.time())
-            individual_predictions_name = self._promised_name(name)
+            individual_predictions_name = self._random_promised_name(name)
+            set_and_expire_pipe.delete(individual_predictions_name)
             set_and_expire_pipe.zadd(name=individual_predictions_name, mapping=predictions, ch=True)  # (1)
             promise_ttl = max(self.DELAYS) + self._DELAY_GRACE
             set_and_expire_pipe.expire(name=individual_predictions_name, time=promise_ttl)   # (2)
@@ -977,8 +993,32 @@ class Rediz(RedizConventions):
 
             # Execute pipeline ... should not fail (!)
             execut = set_and_expire_pipe.execute()
-            anticipated_execut = [self.NUM_PREDICTIONS]*len(delays) + [ self.NUM_PREDICTIONS, True ] + [ 1, True, True ]*len(delays)
-            success = all( actual==anticipate for actual, anticipate in zip(execut, anticipated_execut) )
+            anticipated_execut = [1, self.num_predictions]*len(delays) + [ self.num_predictions, True ] + [ 1, True, True ]*len(delays)
+
+            def _close(a1,a2):
+                return a1==a2 or ( isinstance(a1,int) and a1>20 and ((a1-a2)/self.num_predictions)<0.05 )
+
+            success = all( _close(actual,anticipate) for actual, anticipate in zip(execut, anticipated_execut) )
+
+            if success:
+                # Confirmation log
+                cnfrms = self.confirms_name(write_key=write_key)
+                confirmation = {'type':'prediction','success':success,'name':name,'delays':delays,'values':values[:5]}
+                confirm_pipe = self.client.pipeline(transaction=False)
+                confirm_pipe.lpush(cnfrms, json.dumps(confirmation))
+                confirm_pipe.expire(cnfrms, self.CONFIRMS_TTL)
+                confirm_pipe.ltrim(cnfrms, start=0, end=self.CONFIRMS_MAX)
+                confirm_pipe.execute(raise_on_error=True)
+            else:
+                rrs = self.errors_name(write_key=write_key)
+                error_report = {'type': 'prediction', 'success': success, 'name': name, 'delays': delays,
+                                'values': values[:5],'anticipated_execut':anticipated_execut,'actual_execut':execut}
+                error_pipe = self.client.pipeline(transaction=False)
+                error_pipe.lpush(rrs, json.dumps(error_report))
+                error_pipe.expire(rrs, self.ERROR_TTL)
+                error_pipe.ltrim(rrs, start=0, end=self.ERROR_LIMIT)
+                error_pipe.execute(raise_on_error=True)
+
             return success
         else:
             # TODO: Log failed prediction attempt to write_key log
@@ -1054,20 +1094,20 @@ class Rediz(RedizConventions):
                             # Record keeping
                             rescaled_amount = budget * float(amount)
                             pipe.hincrbyfloat(name=self._BALANCES, key=recipient, amount=rescaled_amount)
-                            write_code = RedizConventions.hash(write_key)
-                            recipient_code = RedizConventions.hash(recipient)
-                            pipe.zincrby(name=self.performance_name(name=name), value=recipient_code, amount=rescaled_amount)              # Public performance on stream
-                            pipe.zincrby(name=self.performance_name(name=name,delay=delay), value=recipient_code, amount=rescaled_amount)  # Public performance on stream
-                            pipe.zincrby(name=self.performance_name(name=recipient), value=name, amount=rescaled_amount)                   # Private performance across streams
+                            from muid import shash
+                            write_code = muid.shash(write_key)
+                            recipient_code = muid.shash(recipient)
+                            # Leaderboards: overall, by stream, by stream and horizon
+                            for lb in [self.leaderboard_name(), self.leaderboard_name(name=name), self.leaderboard_name(name=name,delay=delay) ]:
+                                pipe.zincrby(name=lb, value=recipient_code, amount=rescaled_amount)
+                            # Performance: custom report for the recipient
+                            pipe.zincrby(name=self.performance_name(write_key=recipient), value=self.performance_key(name=name,delay=delay), amount=rescaled_amount)
+                            # Transactions logs:
                             transaction_record = {"settlement_time":str(datetime.datetime.now()),"stream": name, "delay":delay, "stream_owner_code":write_code,"recipient_code":recipient_code, "amount": rescaled_amount}
-                            key_trans_name = self.transactions_name(write_key=recipient)
-                            name_trans_name = self.transactions_name(name=name)
-                            key_name_trans_name = self.transactions_name(write_key=recipient,name=name)
-                            name_delay_trans_name = self.transactions_name(write_key=recipient, name=name, delay=delay )
-                            all_trans_name = self.transactions_name(write_key=None,name=None,delay=None)
-                            for tn in [all_trans_name, key_trans_name, name_trans_name, key_name_trans_name,name_delay_trans_name]:
-                                pipe.xadd(name=tn, fields=transaction_record )
-                                pipe.expire(name=tn,      time=self._TRANSACTIONS_TTL)
+                            log_names = [ self.transactions_name(), self.transactions_name(write_key=recipient ),  self.transactions_name(write_key=recipient, name=name ), self.transactions_name(write_key=recipient, name=name, delay=delay ) ]
+                            for ln in log_names:
+                                pipe.xadd(name=ln ,   fields=transaction_record )
+                                pipe.expire(name=ln,  time=self._TRANSACTIONS_TTL)
         settle_exec = pipe.execute()
 
         # Derived markets ... z's for 1-d, 2-d, 3-d market-implied z-scores and z-curves
@@ -1114,11 +1154,11 @@ class Rediz(RedizConventions):
         game_payments = Counter(dict((p, -1.0) for p in participant_set))
 
         if len(rewarded_scenarios) == 0:
-            carryover = Counter({self._RESERVE: 1.0 * pool / self.NUM_PREDICTIONS})
+            carryover = Counter({self._RESERVE: 1.0 * pool / self.num_predictions})
             game_payments.update(carryover)
         else:
             winners = [self._scenario_owner(ticket) for ticket in rewarded_scenarios]
-            reward = (1.0 * pool / self.NUM_PREDICTIONS) / len(winners)  # Could augment this to use kernel or whatever
+            reward = (1.0 * pool / self.num_predictions) / len(winners)  # Could augment this to use kernel or whatever
             payouts = Counter(dict([(w, reward * c) for w, c in Counter(winners).items()]))
 
             game_payments.update(payouts)
@@ -1209,12 +1249,6 @@ class Rediz(RedizConventions):
     #            Implementation  (getters)
     # --------------------------------------------------------------------------
 
-    def _get_balance_implementation(self, write_key=None, write_keys=None, aggregate=True):
-        write_keys = write_keys or [ write_key ]
-        balances   = self.client.hmget(self._BALANCES, *write_keys)
-        fixed_balances = [ float( b or "0") for b in balances ]
-        return np.nansum( fixed_balances ) if aggregate else fixed_balances
-
     def _get_lagged_implementation(self, name, with_times, with_values, to_float, start=0, end=None, count=100 ):
         count = count or self.LAGGED_LEN
         end = end or start + count
@@ -1294,9 +1328,9 @@ class Rediz(RedizConventions):
     def _get_prefixed_implementation(self, prefixed_name):
         """ Interpret things like  delayed::15::air-pressure.json cdf::70::air-pressure.json etc """
         assert self.SEP in prefixed_name, "Expecting prefixed name with "+self.SEP
-        parts = prefixed_name.lower().split(self.SEP)
+        parts = prefixed_name.split(self.SEP)
         if len(parts)==2:
-            ps = parts[0]+self.SEP
+            ps = (parts[0]+self.SEP).lower()
             if ps == self.BACKLINKS:
                 data = self.get_backlinks(name=parts[-1])
             if ps == self.SUMMARY:
@@ -1320,7 +1354,7 @@ class Rediz(RedizConventions):
             elif ps == self.HISTORY:
                 data = self.get_history(name=parts[-1])
             elif ps == self.BALANCE:
-                data = self.get_balance(write_key=parts[-1])
+                data = self.get_performance(write_key=parts[-1])
             elif ps == self.TRANSACTIONS:
                 data = self.get_transactions(write_key=parts[-1])
             else:
@@ -1389,13 +1423,17 @@ class Rediz(RedizConventions):
         return history
 
 
-    def _get_transactions_implementation(self, min, max, count, write_key=None, name=None ):
-        trans_name=self.transactions_name(write_key=write_key,name=name)
-        return self.client.xrevrange(name=trans_name, min=min, max=max, count=count )
+    def _get_transactions_implementation(self, min, max, count, write_key=None, name=None, delay=None ):
+        trnsctns =self.transactions_name(write_key=write_key,name=name,delay=delay)
+        return self.client.xrevrange(name=trnsctns, min=min, max=max, count=count )
 
-    def _get_performance_implementation(self,name,delay, count):
+    def _get_leaderboard_implementation(self, name, delay, count):
         pname = self.performance_name(name=name,delay=delay)
-        return list(reversed(self.client.zrange(name=pname, start=-count-1, end=-1, withscores=True)))
+        leaderboard = list(reversed(self.client.zrange(name=pname, start=-count-1, end=-1, withscores=True)))
+        return leaderboard
+
+    def _get_performance_implementation(self, write_key ):
+        return self.client.hgetall(self.performance_name(write_key=write_key ))
 
     def _get_links_implementation(self, name, delay=None, delays=None):
         """ Set of outgoing links created by name owner """
@@ -1424,6 +1462,12 @@ class Rediz(RedizConventions):
     def _get_samples_implementation(self, name, delay=None, delays=None, obscure=True):
         return self._get_distribution( namer=self._samples_name, name=name, delay=delay, delays=delays, obscure=obscure )
 
+    def _get_index_implementation(self):
+        ownership = list(self.client.hgetall(self._OWNERSHIP))
+        obscured = [(name, muid.animal(key)) for name, key in ownership.items()]
+        obscured.sort(key=lambda t: len(t[1]))
+        return obscured
+
     def _get_distribution(self, namer, name, delay=None, delays=None, obscure=True):
         """ Get predictions or samples and obfuscate (hash) the write keys """
         singular = delays is None
